#!/usr/bin/env python3
import os, sys, argparse

#for absolute imports to work in script mode, we need to import from the root folder
opath = os.path.abspath(".")
abspath = os.path.abspath(__file__)
abspath = abspath[:abspath.rfind('/') + 1]
os.chdir(abspath)
abspath = os.path.abspath("../..")
sys.path.insert(0, abspath)

from burst.config import menus, summary, configurers
from burst.config.config_file_utils import get_config, write_config

os.chdir(opath)

# CONFIG_FILE = '~/.burst/config.yml'

def main():

    args = parse_arguments()

    config = get_config(args.config_path)
    print('Welcome to the burst tool configuration!')
    if args.main_choice in [None, 'summary']:
        summary.all(config)

    main_selection = args.main_choice if args.main_choice else menus.main_menu()
    if main_selection == 'compute':
        second_selection = args.second_choice if args.second_choice else menus.main_service_menu('compute')
        if second_selection == 'add':
            alias, creds = configurers.new_compute(aws_path=args.aws_path)
            configurers.check_existance(config, 'compute', alias)
            config['compute']['configurations'][alias] = creds
            print(f'{alias} has been added to compute')
        elif second_selection == 'remove':
            remove_alias = configurers.remove_service(config, 'compute', remove_alias=args.third_choice)
            del config['compute']['configurations'][remove_alias]
            print(f'{remove_alias} has been removed from compute')

    elif main_selection == 'storage':
        second_selection = args.second_choice if args.second_choice else menus.main_service_menu('storage')
        if second_selection == 'add':
            alias, creds = configurers.new_storage(aws_path=args.aws_path)
            config['storage']['configurations'][alias] = creds
            print(f'{alias} has been added to storage')
        elif second_selection == 'remove':
            remove_alias = configurers.remove_service(config, 'storage', remove_alias=args.third_choice)
            del config['storage']['configurations'][remove_alias]
            print(f'{remove_alias} has been removed from storage')

    elif main_selection == 'default':
        second_selection = args.second_choice if args.second_choice else menus.default_service_menu()
        if second_selection == 'compute':
            settings = configurers.set_default(config, 'compute')
            config['compute']['settings'] = settings
        elif second_selection == 'storage':
            settings = configurers.set_default(config, 'storage')
            config['storage']['settings'] = settings

    else:
        return 0

    if len(config.get('compute', {}).get('configurations', {})) == 1 and not config['compute']['settings']:
        settings = configurers.set_default(config, 'compute', list(config['compute']['configurations'])[0])
        config['compute']['settings'] = settings

    if len(config.get('storage', {}).get('configurations', {})) == 1 and not config['storage']['settings']:
        settings = configurers.set_default(config, 'storage', list(config['storage']['configurations'])[0])
        config['storage']['settings'] = settings

    write_config(config, args.config_path)


def parse_arguments():

    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('main_choice', choices=['compute', 'storage', 'default', 'summary'], nargs='?')
    parser.add_argument('second_choice', nargs='?')
    parser.add_argument('third_choice', nargs='?')

    parser.add_argument('--aws_path', '-a', default='~/.aws', help='Specify location of AWS credentials (default "~/.aws")')
    parser.add_argument('--config_path', default="~/.burst.config.yml", help='Specify location of config file (default "~/.burst.config.yml")')

    args = parser.parse_args()

    if args.main_choice in ['compute', 'storage'] and args.second_choice not in ['add', 'remove', None]:
        raise Exception("Can only specify 'add' or 'remove' when configuring compute")
    elif args.main_choice == 'default' and args.second_choice not in ['compute', 'storage', None]:
        raise Exception("Can only specify 'compute' or 'storage' when configuring defaults")

    return args


if __name__ == '__main__':
    main()



import os, sys, time, argparse, json
import yaml
from pprint import pprint
from libcloud.compute.types import Provider
from libcloud.compute.providers import get_driver
from libcloud.compute.base import NodeAuthSSHKey
from libcloud.common.google import ResourceNotFoundError
from easydict import EasyDict as dictobj

from burst.verbos import vprint

config = dictobj()

#
# for now providers are EC2 or GCE
#
def init(conf = None):
    #init is a one-time thang
    if 'driver' in config:
        return

    if conf == None:
        conf = {}

    yam = conf.get('configfile', os.environ['HOME'] + "/.burst/config.yml")

    if os.path.exists(yam):
        #FIXME: check for local overriding .burst
        f = open(yam)
        yconf = yaml.load(f, Loader=yaml.FullLoader)
        f.close()
        if 'compute_config' in conf:
            compute_config = conf['compute_config']
        else:
            compute_config = yconf['compute']['settings']['default_compute']
                                                                                    #this got a bit strained. sorry
        storage_config = None
        if 'storage_config' in conf:                                                #if storage_config passed in, use
            storage_config = conf['storage_config']
        else:
            if 'storage' in yconf:                                                  #otherwise check in config.yml
                storage_config = yconf['storage']['settings']['default_storage']
        if storage_config:                                                          #if it exists,
            storage = yconf['storage']['configurations'][storage_config]            #use it
            storage['config'] = storage_config                                      #and store the config name too
        yconf = yconf['compute']['configurations'][compute_config]
        yconf.update(yconf['settings'])   #easier to deal with all attributes at top level
        yconf['compute_config']=compute_config
        if storage_config:                                                          #if specified,
            yconf['storage'] = storage                                              #pull storage to top level for ease

    else:
        vprint ("config.yml not found")
        yconf = {}          #dummy yconf

    if 'provider' in conf:
        config.provider = conf['provider']
    else:
        if 'provider' in yconf:
            config.provider = yconf['provider']
        else:
            raise Exception("Configuration file %s not available. Try running:\nburst --configure" % yam)

    for param in ['access', 'secret', 'region', 'project', 'default_image', 'default_size', 'default_gpu_image',
                  'default_gpu_size', 'default_gpu', 'storage', 'compute_config']:
        if param in conf:
            config[param] = conf[param]
        else:
            config[param] = yconf.get(param, None)

    cls = get_driver(Provider[config.provider])

    if config.provider == 'EC2':
        config.driver = cls(config.access, config.secret, region=config.region)

    elif config.provider == 'GCE':
        if hasattr(config.secret, 'lower'):         #string points to key file
            privkeypath = config.secret
            config.raw_secret = config.secret
        else:                                       #if dict, create key file
            config.raw_secret = "%s.json" % config.secret['private_key_id']
            privkeypath = "%s/.burst/%s.json" % (os.path.expanduser("~"), config.secret['private_key_id'])
        if not os.path.exists(privkeypath):
            fp =  open(privkeypath, 'w')
            json.dump(config.secret, fp)
            fp.close()
        config.driver = cls(config.access, privkeypath, datacenter=config.region, project=config.project)
    else:
        vprint ("ERROR: unknown cloud provider", config.provider)

def get_config():
    return config

def get_server(url=None, uuid=None, name=None, conf = None):
    init(conf)
    nodes = config.driver.list_nodes()
    if url:
        node = [x for x in nodes if url in x.public_ips and x.state != 'terminated']
    elif uuid:
        node = [x for x in nodes if x.uuid.find(uuid)==0 and x.state != 'terminated']
    elif name:
        node = [x for x in nodes if x.name==name and x.state != 'terminated']
    else:
        return "error: specify url, uuid, or name"
    return node[0] if node else None

def get_server_state(srv):
    nodes = config.driver.list_nodes()                      #need to refresh node to get state
    node = [x for x in nodes if x.uuid.find(srv.uuid)==0]
    if node:
        return node[0].state
    vprint ("Cannot find server to determine state; assuming terminated")
    return 'terminated'

def get_server_size(srv):
    if config.provider=='EC2':
        return srv.extra['instance_type']
    elif config.provider=='GCE':
        typ = srv.extra['machineType']
        i = typ.rfind('/')
        return typ[i+1:]

# not working now that EC2 image == AMI full name
# def get_server_image(srv):
#     if config.provider=='EC2':
#         pprint(srv.extra)
#         return srv.extra['name']
#     elif config.provider=='GCE':
#         return srv.extra['image']

def start_server(srv):
    result = srv.start()
    if not result:
        return "error starting server"
    state = None
    while state != 'running':
        state = get_server_state(srv)
        time.sleep(2)
        vprint ("server state:", state)
    vprint ("Waiting for public IP address to be assigned")
    config.driver.wait_until_running([srv])
    vprint("Public IP's:", srv.public_ips)
    while len(srv.public_ips)==0 or srv.public_ips.count(None) == len(srv.public_ips): #Really? Google? [None]????
        # srv = config.driver.list_nodes(ex_node_ids=[srv.id])[0]
        srv = get_server(uuid=srv.uuid)       #seems necessary to refresh to update state
        vprint("Public IP's:", srv.public_ips)
        time.sleep(5)
    return srv

#
# fill in default values for size & image
#
def fix_size_and_image(size, image):
    if image=="DEFAULT_IMAGE":
        image = config.default_image

    if size=="DEFAULT_SIZE":
        size = config.default_size

    if image=="DEFAULT_GPU_IMAGE":
        image = config.default_gpu_image

    if size=="DEFAULT_GPU_SIZE":
        size = config.default_gpu_size
    return size, image

def launch_server(name, size=None, image=None, pubkey=None, conf = None, user=None, gpus=None):
    init(conf)
    size, image = fix_size_and_image(size, image)
    image_full_path = image
    if config.provider=='EC2':
        images = config.driver.list_images(ex_filters={'name': image})
    elif config.provider=='GCE':
        #note: GCE libcloud driver list_images is hella borke, list is incomplete so...
        images = []
        for proj in ["deeplearning-platform-release", "ubuntu-os-cloud"]:
            try:
                im = config.driver.ex_get_image(image, ex_project_list=proj)
                images = [im]
                break
            except ResourceNotFoundError:
                pass
    else:
        ims = config.driver.list_images()
        images = [x for x in ims if x.name == image]
    if not images:
        raise Exception("Image %s not found" % image)
    image = images[0]

    sizes = [x for x in config.driver.list_sizes() if x.name == size]
    if not sizes:
        raise Exception("Instance size %s not found" % size)
    size = sizes[0]
    vprint ("Launching instance image=%s, id=%s, session=%s, type=%s ram=%s disk=%s" % (image_full_path, image.id, name, size.id, size.ram, size.disk))

    if pubkey:
        if config.provider == 'EC2':                #Everybody makes it up
            auth = NodeAuthSSHKey(pubkey)
            node = config.driver.create_node(name, size, image, auth=auth)
        elif config.provider == 'GCE':
            meta = {
                'items': [
                    {
                        'key': 'sshKeys',
                        'value': '%s: %s' % (user, pubkey)
                    }
                ]
            }
            if gpus:
                vprint ("Launching with GPU")
                node = config.driver.create_node(name, size, image, ex_metadata=meta, ex_accelerator_type=config.default_gpu,
                                             ex_accelerator_count=1, ex_on_host_maintenance="TERMINATE")
            else:
                vprint ("Launching without GPU")
                node = config.driver.create_node(name, size, image, ex_metadata=meta)
        else:
            raise Exception("Unsupported clown provider: %s" % config.provider)
    else:
        node = config.driver.create_node(name, size, image)
    vprint ("Waiting for public IP address to be active")
    config.driver.wait_until_running([node])
    while len(node.public_ips)==0:
        # node = config.driver.list_nodes(ex_node_ids=[node.id])[0] #refresh node -- is this really necessary
        node = get_server(uuid=node.uuid)       #seems necessary to refresh to update state
        vprint("Public IP's:", node.public_ips)
        time.sleep(5)
    return node

def stop_server(srv):
    result = srv.stop_node()
    if not result:
        return "error stopping server"
    state = None
    while state != 'stopped':
        state = get_server_state(srv)
        time.sleep(2)
        vprint ("server state:", state)
    return "success"

def terminate_server(srv):
    result = config.driver.destroy_node(srv)
    if not result:
        return "error terminating server"
    state = None
    while state != 'terminated':
        state = get_server_state(srv)
        time.sleep(2)
        vprint ("server state:", state)
    return "success"

def list_servers(name, conf = None, terminated=True):
    init(conf)
    ret = []
    nodes = config.driver.list_nodes()
    for x in nodes:
        x = get_server(uuid=x.uuid)       #seems necessary to refresh to update state
        if not x:
            continue
        # print ("DBG", terminated, x.state)
        if (not terminated) and (x.state=='terminated'): #don't include terminated
            continue
        if x.name==name:
            ret.append([x])
            img = x.extra['image_id'] if config.provider == 'EC2' else x.image
            if img == config.default_image:
                img += " (default_image, no gpu)"
            elif img == config.default_gpu_image:
                img += " (default_gpu_image)"
            s = "IMAGE: %s STATE: %s IP's: %s ID: %s/%s" %(img, x.state, x.public_ips, config.provider, x.id)
            ret[-1].append(s)
    return ret


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--url")
    parser.add_argument("--uuid")
    parser.add_argument("--name")
    args, unknown = parser.parse_known_args()

    n = get_server(args.url, args.uuid, args.name)
    pprint (n)


#!/usr/bin/env python3
import os, sys, argparse, subprocess, time, traceback, json, getpass
#
# the BDFL does not admire scripts which are also importable modules
# well, frack him -- this is how we roll
#
#so absolute imports work in script mode, we need to import from the parent folder
opath = os.path.abspath(".")
abspath = os.path.abspath(__file__)
abspath = abspath[:abspath.rfind('/') + 1]
os.chdir(abspath)
abspath = os.path.abspath("..")
sys.path.insert(0, abspath)

from burst.lcloud import *
from burst.runrun import run
from burst.version import version
from burst.verbos import set_verbosity, get_verbosity, vprint, vvprint, v0print, get_piper, get_rsync_v, get_dockrunflags

os.chdir(opath)

DEFAULT_IMAGE = "burst_image" #FIXME: should be unique to folder structure
MONITOR_IMAGE = "burstableai/burst_shutdown:latest"
# DOCKER_REMPORT = "2376"
# DOCKER_REMOTE = "localhost:"+DOCKER_REMPORT

def burst(args, sshuser=None, url=None, uuid=None, burst_user=None, gpus = "", ports=None, stop=False,
          image=None, size=None, pubkey=None, dockerfile="Dockerfile",
          cloudmap="", dockerdport=2376, conf=None):
    tunnel = None
    try:
        if not os.path.exists(dockerfile):
            raise Exception("Dockerfile not found")
        if not os.path.exists(".dockerignore"):
            raise Exception("""

.dockerignore file not found. Burst requires a .dockerignore to avoid sending excess data to docker build.
Because the working directory is rsync'd to the remote host, you typically only need to send the Dockerfile
and files that are referred to (such as requirements.txt) to the build daemon.

#Template .dockerignore
#Default to ignoring everything:
**
#exceptions (These will be sent to the docker daemon for building):
!/Dockerfile*
!requirements.txt
""")

        #if url specified, split into user & IP
        if url:
            if not sshuser:
                sshuser, url = url.split('@')

        #launch, restart, or reconnect to node
        node = None

        #unless running --local:
        if url or uuid or burst_user:

            #if server does not exist, launch a fresh one
            fresh = False
            restart = False
            node = get_server(url=url, uuid=uuid, name=burst_user, conf=conf)
            if burst_user and not node:
                node = launch_server(burst_user, pubkey=pubkey, size=size, image=image, conf=conf, user=sshuser, gpus=gpus)
                fresh = True
                restart = True
            if node:

                #if stopped, restart
                if node.state.lower() != "running":
                    restart = True
                    vprint ("Starting server")
                    node = start_server(node)

                #by now we must have a public IP address
                url = node.public_ips[0]

                #wait for ssh daemon to be ready
                vprint ("Waiting for sshd")
                cmd = ["ssh", "-o StrictHostKeyChecking=no", "-o UserKnownHostsFile=/dev/null", "-o LogLevel=error", "{0}@{1}".format(sshuser, url), "echo", "'sshd responding'"]
                vvprint(cmd)
                good = False
                for z in range(10, -1, -1):
                    ret = run(cmd, timeout=15)
                    if ret[0].strip()[-15:]=='sshd responding':
                        good = True
                        break
                    vprint ("sshd not responding; %d attempts left" % z)
                    if z:
                        time.sleep(5)
                if not good:
                    raise Exception("error in ssh call: %s" % ret[0].strip())
                vvprint ("SSH returns -->%s|%s<--" % ret)
            else:
                raise Exception("Error: node not found")

        #we have a url unless running --local:
        if url:

            #if just launched, install docker
            if fresh:
                print("Configuring Docker")
                # 'sudo apt-get -y update; sudo apt-get -y install docker.io; ' \ #images have docker installed
                cmd = 'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error {0}@{1} ' \
                      '"sudo usermod -a -G docker ubuntu; ' \
                      'sudo systemctl unmask docker; sudo service docker start"'.format(sshuser, url)
                vvprint(cmd)
                os.system(cmd)

            vprint ("Connecting through ssh")

            #set up ssh tunnel mapping docker socket, ports
            host_port_args = []
            docker_port_args = ""
            if ports:
                for pa in ports:
                    if ':' in pa:
                        local_port, remote_port = pa.split(':')
                    else:
                        remote_port = local_port = pa
                    docker_port_args += " -p {0}:{0}".format(remote_port)
                    host_port_args.append("-L {0}:localhost:{1}".format(local_port, remote_port))
            # print ("PORTS: |%s|%s|" % (docker_port_args, host_port_args)); exit()
            remote = "-H localhost:%s" % dockerdport
            ssh_args = ["ssh", "-o StrictHostKeyChecking=no", "-o UserKnownHostsFile=/dev/null",
                        "-o LogLevel=error", "-NL", "{0}:/var/run/docker.sock".format(dockerdport), "{0}@{1}".format(sshuser, url)]
            for arg in host_port_args:
                ssh_args.insert(3, arg)
            vvprint (ssh_args)
            tunnel = subprocess.Popen(ssh_args)
            time.sleep(2)

            #path = absolute working directory on host
            relpath = os.path.abspath('.')[len(os.path.expanduser('~')):]
            relpath = "/_BURST" +  relpath.replace('/', '_') #I can exlain
            locpath = os.path.abspath('.')
            path = "/home/{0}{1}".format(sshuser, relpath)

            #part of check to see if docker is installed and running
            cmd = ["docker", "{0}".format(remote), "ps", "--format", '{{json .}}']
            vvprint (cmd)
            out = run(cmd)
            vvprint("PS returns -->%s|%s<--" % out)

            # #if any docker shutdown processes are running, kill
            # if out[0].strip():
            #
            #     #kill any pending shutdown containers FIXME: should only fix shutdowns associated with session
            #     kills = []
            #     for x in out[0].split("\n"):
            #         if x:
            #             try:
            #                 j = json.loads(x)
            #             except:
            #                 raise Exception("ERROR in Docker check (is Docker installed?): %s" % x)
            #             Command = j['Command']
            #             if Command.find("burst --stop") < 2:
            #                 kills.append(j['ID'])
            #     if kills:
            #         vprint ("Killing shutdown processes:", kills)
            #         cmd = "docker {0} stop {1} {2} &".format(remote, " ".join(kills), get_piper())
            #         vvprint (cmd)
            #         os.system(cmd)

            #if restarted (including fresh launch), start monitor docker process

            if restart:
                vprint ("Starting monitor process for shutdown++")
                conf = get_config()
                if conf.provider == "GCE":
                    secret = ".burst/" + conf.raw_secret
                else:
                    secret = conf.secret
                # print("SECRET 1:", secret)
                cmd = f"docker {remote} run --rm {get_dockrunflags()} -v {path}:/home/burst/work" \
                      f" {MONITOR_IMAGE} burst-monitor" \
                      f" --ip {url} --access {conf.access} --provider {conf.provider} {get_piper()}" \
                      f" --secret={secret} --region {conf.region} {('--project ' + conf.project) if conf.project else ''}"
                vvprint (cmd)
                vvprint ("Shutdown process container ID:")
                os.system(cmd)

            #prepare to build docker container
            vprint ("Removing topmost layer")        #to avoid running stale image
            cmd = ["docker", "{0}".format(remote), "rmi", "--no-prune", DEFAULT_IMAGE]
            vvprint (cmd)
            out, err = run(cmd)
            vvprint (out)
            size, image = fix_size_and_image(size, image)
            if size and size != get_server_size(node):                      #FIXME
                raise Exception("Cannot change size (instance type) -- need to re-launch")

            # get_server_image is broken, need to prompt better here
            # if image and image != get_server_image(node):
            #     raise Exception("FIXME: cannot change host image -- need to terminate & re-launch server")

            vprint ("burst: name %s size %s image %s url %s" % (node.name, size, image, url))

            #if using cloud storage (s3 etc), set up config & auth for rclone
            if cloudmap:
                if remote:
                    stor = get_config()['storage']
                    if stor['provider'] == 'GCS':
                        #create a keyfile & point to it
                        srvacctf = ".rclone_key_%s.json" % stor['settings']['private_key']['private_key_id']
                        f = open(srvacctf, 'w')
                        json.dump(stor['settings']['private_key'], f)
                        f.close()
                        stor['settings']['service_account_file'] = srvacctf

                    # build  & save rclone.conf
                    s = f"[{stor['config']}]\n"
                    for k, v in stor.items():
                        if k != 'settings':
                            s += f"{k} = {v}\n"
                    for k, v in stor['settings'].items():
                        s += f"{k} = {v}\n"
                    f = open(".rclone.conf", 'w')
                    f.write(s)
                    f.close()

            #sync local working data to host
            rsync_ignore_path = os.path.abspath("./.burstignore")
            if not os.path.exists(rsync_ignore_path):
                vprint("creating empty .burstignore")
                os.system("touch .burstignore")
            cmd = 'rsync -rltzu{4} --exclude-from {5} -e "ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error" {0}/. {3}@{1}:{2}/'.format(locpath,
                                        url, path, sshuser, get_rsync_v(), rsync_ignore_path)
            vprint ("Synchronizing project folders")
            vvprint (cmd)
            os.system(cmd)

            # if get_config().provider == 'GCE':
            #     # sync service acct creds (for shutdown)
            #     cmd = 'rsync -rltzu{4} --relative -e "ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error" {0}/./.burst/{5} {3}@{1}:{2}/'.format(os.path.expanduser('~'),
            #                             url, path, sshuser, get_rsync_v(), get_config().raw_secret)
            #     vprint("Synchronizing credentials for shutdown")
            #     vvprint (cmd)
            #     os.system(cmd)

        else:
            vprint ("burst: running locally")
            remote = ""
            path = os.path.abspath('.')

        #actually build container -- for reals
        vprint ("Building docker container")
        cmd = "docker {1} build . --file {2} -t {0} {3}".format(DEFAULT_IMAGE, remote, dockerfile, get_piper())
        vvprint (cmd)
        os.system(cmd)

        args = " ".join(args)
        gpu_args = "--gpus "+gpus if gpus else ""

        #if mounting storage, add arguments & insert commands before (to mount) and after (to unmount) user-specified args
        cloud_args = ""
        if cloudmap:
            cloud, host = cloudmap.split(":")
            args = f"bash -c 'mkdir -p {host}; rclone mount --vfs-cache-mode writes --vfs-write-back 0 --config .rclone.conf {cloud}: {host} & sleep 3; {args}; umount {host}'"
            cloud_args = " --privileged"

        vprint ("Running docker container")
        cmd = "docker {3} run {4} {5} --rm -ti -v {2}:/home/burst/work {6} {0} {1}".format(DEFAULT_IMAGE,
                                                                                  args, path, remote, gpu_args, docker_port_args, cloud_args)
        #run user-specified args
        vvprint (cmd)
        vprint ("")
        v0print ("---------------------OUTPUT-----------------------")
        sys.stdout.flush()
        os.system(cmd)
        sys.stdout.flush()
        v0print ("----------------------END-------------------------")
        sys.stdout.flush()

        #sync data on host back to local
        if url:
            vprint ("Synchronizing folders")
            cmd = "rsync -rltzu{4} --exclude-from {5} -e 'ssh -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null -o LogLevel=error' '{3}@{1}:{2}/.' {0}/".format(locpath,
                                        url, path, sshuser, get_rsync_v(), rsync_ignore_path)
            vvprint (cmd)
            os.system(cmd)

    except Exception as ex:
        if get_verbosity() >= 256:
            v0print ("--------------------------------")
            traceback.print_exc()
            v0print ("--------------------------------")
        else:
            print ()
        print (ex)

    # if url and node:
    #     # set up shutdown process
    #     if stop == 0:
    #         vprint ("Stopping VM at %s immediately as instructed" % url)
    #         stop_server(node)
    #     else:
    #         vprint ("Scheduling shutdown of VM at %s for %d seconds from now" % (url, stop))
    #         conf = get_config()
    #         if conf.provider == "GCE":
    #             secret = ".burst/" + conf.raw_secret
    #         else:
    #             secret = conf.secret
    #         # print("SECRET 1:", secret)
    #         cmd = f"docker {remote} run --rm {get_dockrunflags()} -v {path}:/home/burst/work {MONITOR_IMAGE} burst" \
    #               f" --verbosity {get_verbosity()} --stop_instance_by_url {url} --delay {stop} --access {conf.access}" \
    #               f" --secret={secret} --region {conf.region} {('--project ' + conf.project) if conf.project else ''}" \
    #               f" --provider {conf.provider} {get_piper()}"
    #         vvprint (cmd)
    #         vvprint ("Shutdown process container ID:")
    #         os.system(cmd)

    if tunnel:
        tunnel.kill()

# #
# # Note this function is typically called by the shutdown process so it does
# # not share scope with most of what burst does
# #
# def stop_instance_by_url(url, conf):
#     vprint ("STOP instance with public IP", url)
#     # print ("DEBUG", os.path.abspath('.'), conf.secret)
#     node = get_server(url=url, conf=conf)
#     if not node:
#         vprint ("No active instance found for IP", url)
#     else:
#         vprint ("shutting down node %s" % node)
#         stop_server(node)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__, add_help=False)
    parser.add_argument("command", nargs='?',                   help="Command to run on remote server")
    parser.add_argument("--configure", action="store_true",     help="Interactive configuration")
    parser.add_argument("--configfile",                         help="override default config.yml")
    parser.add_argument("--build", action="store_true",         help="Download and build environment")
    parser.add_argument("--sshuser", default="ubuntu",          help="remote server username")
    parser.add_argument("--local", action="store_true",         help="run on local device")
    parser.add_argument("--list-servers", action="store_true",  help="List all associated remote servers")
    parser.add_argument("--terminate-servers", action="store_true",     help="Terminate associated remote servers")
    parser.add_argument("--version", action="store_true",       help="Print version # & exit")
    parser.add_argument("--storage-config",                     help="override default storage configuration")
    parser.add_argument("--compute-config",                     help="override default compute configuration")
    parser.add_argument("--url",                                help="run on remote server specified by url")
    parser.add_argument("--uuid",                               help="run on remote server specified by libcloud uuid")
    parser.add_argument("--burst_user",                         help="Burst user name; defaults to local username")
    parser.add_argument("--gpus", default='all',                help="defaults to 'all'; can be set to specific gpu or 'none'")
    parser.add_argument("--portmap", "-p", action="append",     help="port mapping; example: -p 8080 or -p 8080:8080")
    parser.add_argument("--access",                             help="libcloud username (aws: ACCESS_KEY)")
    parser.add_argument("--secret",                             help="libcloud password (aws: SECRET)")
    parser.add_argument("--region",                             help="libcloud location (aws: region)")
    parser.add_argument("--project",                            help="GCE project ID")
    parser.add_argument("--provider", default='EC2',            help="GCE, EC2 etc.")
    parser.add_argument("--image",                              help="libcloud image (aws: ami image_id")
    parser.add_argument("--size",                               help="libcloud size (aws: instance_type")
    parser.add_argument("--pubkey",                             help="public key to access server (defaults to ~/.ssh/id_rsa.pub)")
    parser.add_argument("--delay", type=int, default=0,         help="delay command by N seconds")
    parser.add_argument("--verbosity", type=int, default=0,     help="-1: just task output 0: status 1-4: more verbose")
    parser.add_argument("--shutdown", type=int, default=900, nargs='?',   help="seconds before server is stopped (default 15 minutes)")
    # parser.add_argument("--stop_instance_by_url",               help="internal use")
    parser.add_argument("--cloudmap", type=str, default="",     help="map cloud storage to local mount point")
    parser.add_argument("--dockerfile", type=str, default="Dockerfile",    help="Docker file to build the container with if not ./Dockerfile")
    parser.add_argument("--dockerdport", type=int, default=2376, help="local port to map to remote host docker daemon")
    parser.add_argument("--help", action="store_true",          help="Print usage info")

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(1)
    #
    # this got a bit tricky.
    # we want to parse args BEFORE the main command as burst options
    # and pass all args AFTER the main command to the command when it runs remotely
    #
    argv = sys.argv[1:]
    args, unknown = parser.parse_known_args(argv)

    if args.build and args.verbosity < 1:
        set_verbosity(1)
    else:
        set_verbosity(args.verbosity)

    vvprint ("ARGV:", argv)

    if args.command != None:
        i = argv.index(args.command)
    else:
        i = len(argv)

    rexargs = argv[:i]
    vvprint ("REXARGS:", rexargs)
    cmdargs = argv[i:]
    vvprint ("CMDARGS:", cmdargs)
    args = parser.parse_args(rexargs)
    vvprint ("ARGS:", args)

    if args.help:
        parser.print_help()
        sys.exit(1)

    if args.gpus.lower() == 'none':
        args.gpus = None

    #override config credentials on command line: --access implies all must be provided
    if args.access:
        args_compute = dictobj()
        args_compute.access = args.access
        args_compute.secret = args.secret
        args_compute.region = args.region
        args_compute.project = args.project
        args_compute.provider = args.provider
    else:
        burst_conf = {}

        #command line overrides:
        if args.compute_config:
            burst_conf['compute_config'] = args.compute_config

        if args.storage_config:
            burst_conf['storage_config'] = args.storage_config

        if args.configfile:
            burst_conf['configfile'] = args.configfile

    if args.local and (args.uuid or args.url):
        vprint (args)
        parser.error("when specifying --local, do not set --sshuser, --burst_user, --uuid, or --url")
        exit()
    t0 = time.time()
    while time.time()-t0 < args.delay:
        vprint ("%d seconds till action" % (args.delay+.5+t0-time.time()))
        time.sleep(5)

    #set default burst_user if necessary:
    if not (args.burst_user or args.uuid or args.url or args.local or args.version):
        burst_user = getpass.getuser()
        args.burst_user = "burst-" + burst_user
        vprint ("Session: %s" % args.burst_user)

    # #master switch clause. First, stand-alone options
    if args.list_servers:
        init(burst_conf)
        # pprint(get_config())
        cconf = get_config()['compute_config']
        v0print ("-------------------------------------------------------------\nSessions with config %s & user %s:" % (cconf, args.burst_user))
        for _, s in list_servers(args.burst_user, burst_conf):
            print (s)
        v0print ("-------------------------------------------------------------")

    elif args.shutdown == None:
        v0print ("-------------------------------------------------------------")
        count = 0
        for node, s in list_servers(args.burst_user, burst_conf):
            if node.state == "stopped":
                continue
            count += 1
            yes = input("Stopping (warm shutdown) %s, are you sure? (y/n)" % s)
            if yes=='y':
                stop_server(node)
            else:
                print ("Aborted")
        if not count:
            print ("no servers to shut down")
        v0print ("-------------------------------------------------------------")

    elif args.terminate_servers:
        v0print ("-------------------------------------------------------------")
        count = 0
        for node, s in list_servers(args.burst_user, burst_conf, terminated=False):
            count += 1
            yes = input("Terminating %s, are you sure? (y/n)" % s)
            if yes=='y':
                terminate_server(node)
            else:
                print ("Aborted")
        if not count:
            print ("no servers to terminate")
        v0print ("-------------------------------------------------------------")

    elif args.version:
        print ("VERSION:", version)

    elif args.configure:
        if args.configfile:
            yam = args.configfile
        else:
            yam = os.environ['HOME'] + "/.burst/config.yml"
        os.system("burst-config --config_path %s" % yam)

    else:
        #no stand-alone options; do burst for reals
        if args.local:
            pubkey = None
        else:
            if args.pubkey==None:
                try:
                    f=open(os.path.expanduser("~") + "/.ssh/id_rsa.pub")             #FIXME: a bit cheeky
                    pubkey=f.read()
                    f.close()
                except:
                    print ("Public key not found in usual place; please specify --pubkey")
        if args.gpus:
            if args.size == None:
                size = 'DEFAULT_GPU_SIZE'
            else:
                size = args.size
            if args.image == None:
                image = 'DEFAULT_GPU_IMAGE'
            else:
                image = args.image
        else:
            if args.size == None:
                size = 'DEFAULT_SIZE'
            else:
                size = args.size
            if args.image == None:
                image = 'DEFAULT_IMAGE'
            else:
                image = args.image

        if args.build:
            cmdargs = ['echo', 'Build phase 1 success']

        #let's do this thing
        burst(cmdargs, sshuser=args.sshuser, url=args.url, uuid=args.uuid,
              burst_user=args.burst_user, gpus=args.gpus, ports=args.portmap, stop=args.shutdown,
              image=image, size=size, pubkey=pubkey, dockerfile=args.dockerfile, cloudmap=args.cloudmap,
              dockerdport=args.dockerdport, conf = burst_conf)

        if args.build:
            v0print()
            print ("Build phase 2 success")

        vprint ("DONE")
        v0print()



import  subprocess, time, os, sys, re, socket
from blessings import Terminal
bless_term = Terminal()

MAXLINES=100

def print_red(s, **kw):
    print (bless_term.red(s), **kw)

def run(*args, **kw):
    # print ("DBG", args, kw)
    if 'showoutput' in kw:
        showoutput = kw['showoutput']
        # print("showoutput:", showoutput)
        del kw['showoutput']
    else:
        showoutput = False
    if 'timeout' in kw:
        timeout = float(kw['timeout'])
        if showoutput:
            print("running", args[0], "with timeout:", timeout, end=' ')
        del kw['timeout']
    else:
        timeout = 0
    try:
        if not timeout:
            timeout = 10**10
        # print ("args:", args)
        proc = subprocess.Popen(*args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        t0 = time.time()
        out = ""
        complete = False
        while time.time() < t0 + timeout:
            line = proc.stdout.readline().decode('utf8')
            # print ("DBG", type(out), type(line))
            out += line
            i = 0
            while line != "":
                if showoutput:
                    sys.stdout.write(line)
                i += 1
                if i >= MAXLINES:
                    break
                line = proc.stdout.readline().decode('utf8')
                out += line
            if proc.poll() != None:
                complete = True
                #get all output
                line = proc.stdout.readline().decode('utf8')
                out += line
                while line != "":
                    if showoutput:
                        sys.stdout.write(line)
                    sys.stdout.write(line)
                    line = proc.stdout.readline().decode('utf8')
                    out += line
                sys.stdout.flush()
                break
##                sys.stdout.write(".")
##                sys.stdout.flush()
            time.sleep(0.2)
        if not complete:
            proc.kill()

    except subprocess.CalledProcessError as e:
        out = e.output
    return out, complete

#
# run a git command, capture the output
#
def git(cmd, show=False, debug=False):
    if debug:
        print_red ("git %s" % cmd)
    if hasattr(cmd, "lower"):
        cmd = cmd.split()
    out, good = run(["git"] + cmd, showoutput=show)
    if not good:
        err = "ERROR -- git command did not complete"
        print (err, file=sys.stderr)
        out += "\n\n" + err
    return out, not good


def get_branch():
    out, err = git("rev-parse --abbrev-ref HEAD")
    return out.strip()

def get_repo():
    out, err = git("remote -v")
    return out.split()[1]

def get_author():
    return git("log -1 --pretty=format:'%an'")[0].strip().replace("'", "")

def get_username():
    return git("config --get user.name")[0].strip()

def git_status(show=False, debug=False):
    out, err = git("status --porcelain", show=show, debug=debug)
    changes=0
    for row in out.split("\n"):
        row = row.strip()
        if not row:
            continue
        if row[:2] != "??":
            changes += 1
    return changes

import subprocess as sp
from threading import Thread
from queue import Queue, Empty
import time


def test_func(s):
    if not s:
        return ""
    # print ("----------PARSE------------")
    # print (s)
    # print ("~~~~~~~~~~~~~~~~~")
    N = 7
    for L in s.split():
        try:
            N = int(L.strip())
        except:
            pass
    # print ("RESULT:", N)
    # print ("----------/PARSE------------")
    return "BOOM " * N


def stdout_thread(o, q):
    def getchar():
        return o.read(1)

    for c in iter(getchar, b''):
        q.put(c)
    o.close()


def get_sub_stdout(q):
    r = b''
    while True:
        try:
            c = q.get(False)
        except Empty:
            # print ("   EMPTY")
            break
        else:
            # print ("   DATA")
            r += c
    return r

def escape_ansi(line):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    # ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', line)


SLEEPYTIME = .1
SSH_FORCE_TIMEOUT = 30
class runner:
    def __init__(self, cmd):
        self.pobj = sp.Popen(cmd.split(), stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.STDOUT)
        self.q = Queue()
        self.t = Thread(target=stdout_thread, args=(self.pobj.stdout, self.q))
        self.t.daemon = True
        self.t.start()
        self.in_dat = ''
        self.t0 = time.time()

    # Use advanced machine learning algorithms to ascertain if we have a prompt:
    def has_prompt(self, s):  # A proud moment in hell
        # print ("SSS:", s)
        if "\n" in s:
            s = s.split("\n")[-1]
        s = escape_ansi(s.strip())
        i = s.find(self.prompt)
        if i<0  or i>12:
            # print ("FAIL")
            return False
        # print("PROMPT FOUND")
        return True

    #
    # call interact with user input, returns next process text+prompt
    #
    def interact(self, cmd=None, expect=None):
        if cmd != None:                                   #typically None for first interaction to get prompt
                                                          #if '', still need to write to stdin to keep rolling ball
            # print ("===%s==="%cmd)
            self.pobj.stdin.write(bytes(cmd, 'utf-8'))
            self.pobj.stdin.write(b'\n')
            try:
                self.pobj.stdin.flush()
            except:
                return ''
            self.in_dat = cmd

        if expect==None:
            expect=[]
        elif hasattr(expect, "lower"):
            expect = [expect]
        # print ("EXPECT:", expect)
        o_new = get_sub_stdout(self.q).decode('utf8')
        o_dat = o_new
        while not o_new:
            br = False
            for ex in expect:
                # print ("TEST:", ex, o_new, "||", ex in o_new, "|||")
                if ex in o_new:                           #additional triggers to return such as Y/n prompts
                    br = True
                    break
            if br:
                break
            o_new = get_sub_stdout(self.q).decode('utf8')
            o_dat += o_new
            time.sleep(SLEEPYTIME)
        # print ("DBG A")
        # remove echo:
        # if o_dat.find(self.in_dat+"\r\n")==0:
        #     o_dat=o_dat[len(self.in_dat)+2:]
        return o_dat, self.has_prompt(o_dat)

    def first(self):
        o_dat = ""
        t0 = time.time()
        while True:
            # print ("  FIRST:",o_dat)
            if time.time()-t0 > SSH_FORCE_TIMEOUT:
                return o_dat, True
            o_dat += get_sub_stdout(self.q).decode('utf8')
            spl = o_dat.rstrip().split("\n")
            if len(spl) >= 2 and "last login" in spl[-2].lower():
                break
            if "timed out" in spl[-1]:
                return o_dat, True
            time.sleep(SLEEPYTIME)
        # print (o_dat)
        prompt = escape_ansi(spl[-1])
        prompt.replace("\r", ']').strip()
        i = prompt.find(':')
        if i > 0:
            # print ("III:", i)
            prompt = prompt[0:i+1]
        self.prompt = prompt
        print ("PROMPT: >>>%s<<<" % prompt)
        sys.stdout.flush()
        return o_dat, False

    def exit(self):
        self.pobj.stdin.write(bytes('exit', 'utf-8'))
        self.pobj.stdin.write(b'\n')
        time.sleep(2)
        o_new = get_sub_stdout(self.q).decode('utf8')
        print (o_new)
        sys.stdout.flush()

if __name__=="__main__":
    cmd = "ssh -tt -4 localhost"
    # cmd = "echoz foo"
    print (cmd, end="\n\n")
    run = runner(cmd)
    o = run.first()                                  #get initial startup spam + prompt
    print (o)
    run.interact("pwd")
    run.exit()
    print ("DONE")

# if __name__ == "__main__":
#     print("test run.py")
#     cmd = sys.argv[1:]
#     s, err = run(cmd, timeout=10, showoutput=False)
#     print("output----------\n", s)
#     print("end output------")
#     print("completed:", err)



from setuptools import setup, find_packages
from burst.version import version

setup(
    name="burstable",
    version=version,
    py_modules=['config'],
    packages=find_packages(),
    python_requires='>=3.6',
    scripts=['bin/burst', 'bin/burst-config'],
    install_requires=[
        "blessings          >=1.7,      <2",
        "apache-libcloud    >=3.2.0,    <4",
    	"cryptography       >=3.2,      <4",
        "easydict           >=1.9,      <2",
        "PyYAML             >=5.3.1,    <6"
    ],
    include_package_data=True
)



import os, sys

from burst.config import summary
from burst.config.cred_loaders import get_aws_creds
from burst.config.config_file_utils import get_config

abspath = os.path.abspath(__file__)
abspath = abspath[:abspath.rfind('/') + 1]

# The alias used in the template file to set default images/sizes
TEMPLATE_AWS_ALIAS = 'burst_ec2'

def new_compute(aws_path='~/.aws'):

    print('\nSetting up EC2 on AWS')

    config = get_config(file_name='%sconfig_template.yml' % abspath)['compute']['configurations'][TEMPLATE_AWS_ALIAS]
    config['access'], config['settings']['secret'], config['region'] = get_aws_creds(aws_path)
    alias = input('\nPlease enter an alias (name) to reference these credentials: ')
    # ask if this should be default?

    return alias, config


def new_storage(aws_path='~/.aws'):
    print('\nSetting up S3 on AWS')

    config = {'settings': {}}
    config['provider'] = 'AWS'
    config['type'] = 's3'
    config['settings']['access_key_id'], config['settings']['secret_access_key'], config['settings']['region'] = get_aws_creds(aws_path)
    config['settings']['env_auth'] = False
    config['settings']['acl'] = 'private'
    # config['default_mount_folder'] = input("\nSet the default mount folder: ")

    alias = input('\nPlease enter an alias (name) to reference these credentials: ')
    return alias, config


def remove_service(config, service, remove_alias=None):

    if remove_alias is None:
        getattr(summary, service)(config)
        remove_alias = input('Enter name to remove: ')

    if remove_alias not in config[service].get('configurations', []):
        raise Exception("Name entered is not configured")

    return remove_alias


def set_default(config, service, default_service=None):

    settings = {}

    if default_service is None:
        getattr(summary, service)(config)
        default_service = input('\nEnter the name to set for default compute service: ')
        if default_service not in config[service].get('configurations', []):
            raise Exception("Name entered is not configured")
    else:
        print(f'\nSince {default_service} is the only {service} configuration, it will be set to default.')

    if service == 'compute':
        # default_to_gpu = input("Default to GPU (y/n): ")
        # settings['default_to_gpu'] = True if default_to_gpu.lower() in ['y', 'yes'] else False
        settings['default_compute'] = default_service
    elif service == 'storage':
        # sync_policy = input('Enter default sync policy (hit enter to select recommended "lazy"): ')
        # settings['sync_policy'] = sync_policy if sync_policy else 'lazy'
        settings['default_storage'] = default_service

    return settings

def check_existance(config, service, alias):

    if alias in config[service]['configurations']:
        proceed = input(f"Warning - {alias} is already in use. Overwrite? (y/n)> ")

        if proceed.lower() != 'y':
            sys.exit()
